import json
import re
import urllib.request

from bs4 import BeautifulSoup

import database.dbmodels as db
from util.util import get_season


def get_match_ids(year, league):
    web = urllib.request.urlopen("https://understat.com/league/{}/{}".format(league, year))
    soup = BeautifulSoup(web.read(), 'lxml')
    p = re.compile("var\s*datesData\s*=\s*JSON.parse\('(.*)'\)")
    match = None
    for script_tag in soup.find_all("script"):
        if not script_tag.string:
            continue
        match = p.search(str(script_tag.string))
        if match:
            break
    datesData = json.loads(match.group(1).encode("utf-8").decode("unicode_escape"))
    return [d["id"] for d in datesData]


def scrape_match_page(id):
    try:
        page = urllib.request.urlopen("https://understat.com/match/{}".format(id))
        soup = BeautifulSoup(page.read(), 'lxml')
        patterns = {
            "shotsData": re.compile("var\s*shotsData\s*=\s*JSON.parse\('(.*)'\)"),
            "matchInfo": re.compile("match_info\s*=\s*JSON.parse\('(.*)'\)"),
            "rostersData": re.compile("var\s*rostersData\s*=\s*JSON.parse\('(.*)'\)")
        }
        data = dict.fromkeys(["shotsData", "matchInfo", "rostersData"])
        for script_tag in soup.find_all("script"):
            if not script_tag.string:
                continue
            for name, pattern in patterns.items():
                match = pattern.search(str(script_tag.string))
                if match:
                    data[name] = json.loads(match.group(1).encode("utf-8").decode("unicode_escape"))
        return data
    except urllib.error.HTTPError as err:
        if err.code == 404:
            print("Match returned 404, skipping")
            return 0
        else:
            raise


def insert_match_data(matchInfo):
    home, _ = db.Team.get_or_create(name=matchInfo["team_h"])
    away, _ = db.Team.get_or_create(name=matchInfo["team_a"])
    match = db.Match.select().where(db.Match.date == str(matchInfo['date']), db.Match.home == matchInfo['team_h'], db.Match.away == matchInfo['team_a'])
    if match.count() == 0:
        match, _ = db.Match.get_or_create(
            date=matchInfo["date"],
            home=home,
            away=away,
            league=matchInfo["league"],
            season=matchInfo["season"],
            h_goals=matchInfo["h_goals"],
            a_goals=matchInfo["a_goals"],
            h_xg=matchInfo["h_xg"],
            a_xg=matchInfo["a_xg"],
            h_shots=matchInfo["h_shot"],
            a_shots=matchInfo["a_shot"],
            h_onTarget=matchInfo["h_shotOnTarget"],
            a_onTarget=matchInfo["a_shotOnTarget"],
            h_deep=matchInfo["h_deep"],
            a_deep=matchInfo["a_deep"],
            h_ppda=matchInfo["h_ppda"],
            a_ppda=matchInfo["a_ppda"])
    else:
        query = (db.Match.update({db.Match.h_goals: matchInfo['h_goals'],
                 db.Match.a_goals: matchInfo['a_goals'],
                 db.Match.h_xg: matchInfo['h_xg'],
                 db.Match.a_xg: matchInfo['a_xg'],
                 db.Match.h_shots: matchInfo['h_shot'],
                 db.Match.a_shots: matchInfo['a_shot'],
                 db.Match.h_onTarget: matchInfo['h_shotOnTarget'],
                 db.Match.a_onTarget: matchInfo['a_shotOnTarget'],
                 db.Match.h_deep: matchInfo['h_deep'],
                 db.Match.a_deep: matchInfo['a_deep'],
                 db.Match.h_ppda: matchInfo['h_ppda'],
                 db.Match.a_ppda: matchInfo['a_ppda'],
                 })
        .where(db.Match.season == str(get_season(matchInfo['date'][:10], format="%Y-%M-%d")), db.Match.home == matchInfo['team_h'], db.Match.away == matchInfo['team_a']))
        result = query.execute()
        print("Updated match. Result: " + str(result))
    return match


def insert_roster_data(match, rosterData):
    for teamDict in rosterData.values():
        for playerData in teamDict.values():
            player, _ = db.Player.get_or_create(id=int(playerData["player_id"]), name=playerData["player"])
            db.PlayerMatchData.get_or_create(
                player=player,
                match=match,
                team= match.home if playerData["h_a"] == "h" else match.away,
                goals=playerData["goals"],
                own_goals=playerData["own_goals"],
                shots=playerData["shots"],
                xG=playerData["xG"],
                time=playerData["time"],
                position=playerData["position"],
                yellow_card=playerData["yellow_card"],
                red_card=playerData["red_card"],
                key_passes=playerData["key_passes"],
                assists=playerData["assists"],
                xA=playerData["xA"],
                xGChain=playerData["xGChain"],
                xGBuildup=playerData["xGBuildup"],
                positionOrder=playerData["positionOrder"]
            )


def generate_season_dataset(season, league):
    ids = get_match_ids(season, league)
    for id in ids:
        print("scraping match {}".format(id))
        data = scrape_match_page(id)
        if data != 0:
            match = insert_match_data(data["matchInfo"])
            if match != 0:
                insert_roster_data(match, data["rostersData"])
    print("Season done scraping.")


def scrape_all_seasons(reset_db=False, league="EPL"):
    if reset_db:
        db.create_db()
    for season in range(2014, 2018):
        print("Scraping {} season {}".format(league, season))
        generate_season_dataset(season, league)


leagues = ["RFPL"]
for league in leagues:
    scrape_all_seasons(reset_db=False, league=league)
