import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from scipy.signal import savgol_filter

from rlexperiments.betting_env import WeeklyEnvironment, SingleMatchEnvironment
from rlexperiments.dqn_agent import DQNAgent


def plot_history(history, filename=None):
    if len(history) == 0:
        print("No bets!")
        return
    bet_ax = plt.gca()
    bet_ax.set_xlabel("Probability")
    bet_ax.set_ylabel("Odds")
    max_odds = None
    bets_taken = []
    bets_skipped = []
    for bet in history:
        if bet["action"] != 0:
            action = bet["action"] - 1
            for i in range(3):
                prob = bet["state"][i]
                odds = bet["state"][i+3]
                if i == action:
                    bets_taken.append((prob, odds))
                else:
                    bets_skipped.append((prob, odds))
                if max_odds is None or odds > max_odds:
                    max_odds = odds

    bet_ax.plot([b[0] for b in bets_skipped], [b[1] for b in bets_skipped], 'ro', markersize=3)
    bet_ax.plot([b[0] for b in bets_taken], [b[1] for b in bets_taken], 'g^', markersize=3)

    if max_odds is not None:
        x_range = np.arange(1, max_odds, 0.01)
        y = [1 / x for x in x_range]
        bet_ax.plot(y, x_range, "b-")

    legend_elements = [
        Line2D([0], [0], marker='^', markerfacecolor='g', color='w', label="Bets taken"),
        Line2D([0], [0], marker='o', markerfacecolor='r', color='w', label="Bets not taken"),
        Line2D([0], [0], color='b', label="Break even")
    ]
    bet_ax.legend(handles=legend_elements, loc='upper right')

    if filename is not None:
        plt.savefig(f"plots/{filename}.png")
    plt.show()


def test(env, agent, episodes, plot=True, filename=None, aggregate=False):
    old_eps = agent.epsilon
    agent.epsilon = 0 # TURN OFF EXPLORATION
    bet_history = []
    actions = [0, 0, 0, 0]
    total_steps = 0
    total_score = 0
    num_bankrupt = 0
    sum_end_bankroll = 0
    max_money = -float('inf')
    min_money = float('inf')
    for e in range(episodes):
        state = env.reset()
        state = np.reshape(state, [1, agent.state_size])
        accumulated_reward = 0
        if not aggregate:
            bet_history = []
        for time in range(500):
            action = agent.act(state)
            actions[action] += 1
            next_state, reward, done, info = env.step(action)
            bet_history.append({
                "action": action,
                "state": state[0],
                "win": info["profit"] > 0,
                "money": next_state[-1]
            })
            reward = reward if not done else -10
            accumulated_reward += reward
            next_state = np.reshape(next_state, [1, agent.state_size])
            state = next_state
            if done:
                num_bankrupt += 1
                break
        total_steps += time
        print(f"episode: {e+1}/{episodes}, score: {accumulated_reward}, steps: {time}")
        if plot and not aggregate:
            fname = filename + str(e) if filename is not None else filename
            plot_history(bet_history, fname)
        total_score += accumulated_reward
        end_bankroll = bet_history[-1]["money"]
        sum_end_bankroll += end_bankroll
        if end_bankroll > max_money:
            max_money = end_bankroll
        if end_bankroll < min_money:
            min_money = end_bankroll
    num_actions = sum(actions)
    print(f"NB: {actions[0]/num_actions * 100}%, H: {actions[1]/num_actions* 100}%, D: {actions[2]/num_actions * 100}%, A: {actions[3]/num_actions * 100}%")
    print(f"Avg steps: {total_steps / episodes}, Avg score: {total_score / episodes}, Bankrupt: {num_bankrupt} / {episodes}, Avg end money: {sum_end_bankroll / episodes}, End money range: ({min_money}, {max_money})")
    if plot and aggregate:
        plot_history(bet_history, filename + "-aggregate")
    agent.epsilon = old_eps


if __name__ == "__main__":
    EPISODES = 100
    TEST_EPISODES = 5
    EPS_DECAY = 0.999
    env = SingleMatchEnvironment(50, 1, fake_data=True)
    state_size = env.observation_space.shape[0]
    action_size = env.action_space.n
    agent = DQNAgent(state_size, action_size, epsilon_decay=EPS_DECAY, probabilistic_actions=False)
    done = False
    checkpoint_name_offset = 12000
    batch_size = 64
    filename = "fake-data-experiment"
    reward_log_filename = "fake-data-rewards"

    episode_rewards = []
    for e in range(EPISODES+1):
        state = env.reset()
        state = np.reshape(state, [1, state_size])
        accumulated_reward = 0
        acc_optimal = 0
        for time in range(500):
            action = agent.act(state)
            next_state, reward, done, info = env.step(action)
            reward = reward if not done else -10
            accumulated_reward += reward
            acc_optimal += info["optimal"]
            next_state = np.reshape(next_state, [1, state_size])
            agent.remember(state, action, reward, next_state, done)
            state = next_state
            if done:
                break
        episode_rewards.append((accumulated_reward, acc_optimal))
        if e % 10 == 0:
            print("episode: {}/{}, score: {}, e: {:.2}"
                  .format(e, EPISODES, accumulated_reward, agent.epsilon))
        if len(agent.memory) > batch_size:
            agent.replay(batch_size)
        if e % 100 == 0 and e > 0:
            test(env, agent, 1, plot=e % 1000 == 0)
            f = open(f"{reward_log_filename}", "a+")
            for reward in episode_rewards[e-100:e]:
                f.writelines(f"{reward[0]}, {reward[1]}\n")
            f.close()
        if e % 1000 == 0 and e > 0:
            agent.save(f"model-checkpoints/{filename}-checkpoint{e + checkpoint_name_offset}.h5")
            plt.plot(savgol_filter([r[0] for r in episode_rewards], 51, 3), color='b')
            plt.plot(savgol_filter([r[1] for r in episode_rewards], 51, 3), color='g')
            plt.plot()
    agent.save(f"model-checkpoints/{filename}-final.h5")

    f = open(f"{reward_log_filename}")
    rewards = [r.split(", ") for r in f.readlines()]
    f.close()
    episode_rewards = [float(r[0]) for r in rewards]
    smoothed_rewards = savgol_filter(episode_rewards, 51, 3)
    f = open(f"{reward_log_filename}")
    rewards = [r.split(", ") for r in f.readlines()]
    f.close()
    optimal_rewards = [float(r[1]) for r in rewards]
    smoothed_optimal = savgol_filter(optimal_rewards, 51, 3)

    max_len = min(len(optimal_rewards), len(episode_rewards))

    plt.figure()
    plt.plot(episode_rewards[:max_len], color='b', alpha=0.2, label="Actual Rewards")
    plt.plot(smoothed_rewards[:max_len], color='b', label="Smoothed Rewards")
    #plt.plot(optimal_rewards, color='g', alpha=0.2, label="Actual Baseline")
    plt.plot(smoothed_optimal[:max_len], color='g', label="Smoothed Baseline Rewards")
    plt.legend()
    plt.title("Accumulated reward per episode")
    plt.savefig(f"plots/{reward_log_filename}.png")
    plt.show()
    test(env, agent, TEST_EPISODES, plot=True, filename=filename, aggregate=True)